#!/usr/bin/python
import threading, Queue, md5
import urllib2, time, urlparse, os.path
import codecs
import logging, logging.config
import MySQLdb
import mx.DateTime
import RSSParser
import htmlStrip
import AggregatorConfig


# Attempt to read in the configuration for the system
AGG_CONFIG = AggregatorConfig.AggregatorConfig()

loggingConfig = AGG_CONFIG.getValue ('aggregator.loggingConfig')
if (os.path.exists (loggingConfig)):
	logging.config.fileConfig (loggingConfig)
else:
	logging.basicConfig()
	
# This is used to strip out non ascii chars to enable MD5 to work.
ASCII_ENCODER=codecs.lookup('ascii')[0]
dbToUnicode = AGG_CONFIG.dbToUnicode
unicodeToDb = AGG_CONFIG.unicodeToDb

USER_AGENT="TALAggregator/0.4 (+http://www.owlfish.com/software/TALAggregator/)"
	
class Aggregator:
	def __init__ (self):
		self.log = logging.getLogger ("Aggregator")
		self.dbUser = AGG_CONFIG.getValue ('global.user')
		self.dbPass = AGG_CONFIG.getValue ('global.passwd')
		self.dbName = AGG_CONFIG.getValue ('global.db')
		self.pollTime = int(AGG_CONFIG.getValue ('aggregator.sleeptime'))
		self.numWorkers = int(AGG_CONFIG.getValue ('aggregator.numthreads'))
		
		self.db = MySQLdb.connect (db=self.dbName
								 ,user=self.dbUser
								 ,passwd=self.dbPass
								 )
		self.queue = Queue.Queue()
		self.workerList = []
		for workerNum in xrange (0,self.numWorkers):
			self.log.debug ("Creating worker %s" % str (workerNum))
			worker = WorkerThread(self, workerNum)
			self.workerList.append (worker)
			worker.start()
		self.feedsBeingChecked = {}
		self.statusLock = threading.Lock()
			
	def runAggregator (self):
		stmt = "SELECT * FROM agg_feeds WHERE IFNULL(nextCheck,%s) <= %s" 
		cursor = self.db.cursor()
		running = 1
		
		while (running):
			startCheck = time.clock()
			cursor.execute (stmt, (mx.DateTime.gmt(),mx.DateTime.gmt()))
			result = cursor.fetchall()
			for feedToCheck in result:
				self.checkFeed (feedToCheck[0],feedToCheck[1], feedToCheck[2])
			toSleep = startCheck + self.pollTime - time.clock()
			self.log.debug ("Sleeping for %s seconds" % str (toSleep))
			try:
				time.sleep (toSleep)
			except:
				self.log.info ("Sleep interupted, shutting down.")
				for workerNum in xrange (0,self.numWorkers):
					self.queue.put (WorkRequest ("QUIT", None))
				running = 0
		self.log.info ("Waiting for threads to finish.")
		for worker in self.workerList:
			worker.join()
		self.log.info ("All threads finished, stopping.")
				
		
	def checkFeed (self, id, url, checkFrequency):
		self.statusLock.acquire()
		if (self.feedsBeingChecked.has_key (id)):
			self.log.warn ("Feed ID %s (%s) is already in the queue to be checked." % (str (id), url))
		else:
			self.feedsBeingChecked [id] = url
			self.queue.put (WorkRequest ("AGGREGATE", (id, url, checkFrequency)))
		self.statusLock.release()
		
	def feedChecked (self, id):
		self.statusLock.acquire()
		del self.feedsBeingChecked [id]
		self.statusLock.release()
			
	def getQueue (self):
		return self.queue
		
	def getDB (self):
		return MySQLdb.connect (db=self.dbName
							 ,user=self.dbUser
							 ,passwd=self.dbPass
							 )
		
class WorkRequest:
	def __init__ (self, workName, args=()):
		self.workName = workName
		self.args = args
		
	def getWork (self):
		return (self.workName, self.args)
		
	def __str__ (self):
		return "Work %s parameters: %s" % (str (self.workName), str (self.args))
		
class WorkerThread (threading.Thread):
	def __init__ (self, aggregator, workerNum):
		threading.Thread.__init__ (self)
		self.log = logging.getLogger ("Aggregator.WorkerThread.%s" % str (workerNum))
		self.workerNum = workerNum
		self.aggregator = aggregator
		self.queue = aggregator.getQueue()
		self.running = 1
		self.dbConn = aggregator.getDB()
		
	def run (self):
		while (self.running):
			self.log.info ("Worker thread running - waiting for requests")
			requestToProcess = self.queue.get()
			self.log.debug ("Got work request: " + str (requestToProcess))
			# Fast exit criteria - don't start something else if we need to exit
			if (self.running):
				work,params = requestToProcess.getWork()
				if (work == "AGGREGATE"):
					self.aggregate (params)
				elif (work == "QUIT"):
					self.quit()
				else:
					self.log.warn ("Recieved work request %s, unable to handle!" % str (requestToProcess))
		self.log.info ("Worker no longer running, exiting.")
		return 0
		
	def aggregate (self, params):
		id, url, checkFrequency = params
		try:
			feedHandler = FeedHandler (id, url, self.dbConn, self.log)
			feedHandler.getUpdates ()
			feedHandler.setNextCheck (checkFrequency)
		except Exception, e:
			self.log.error ("Exception thrown during feed handling: " + str (e))
		# We have finished handling this url
		self.aggregator.feedChecked (id)
	
	def quit (self):
		# This is the last thing to do before quiting.
		self.running = 0
		self.log.info ("Exiting.")

class FeedHandler:
	def __init__ (self, id, url, db, log):
		self.log = log
		self.id = id
		self.url = url
		self.db = db
		self.cursor = self.db.cursor()
		
	def setNextCheck (self, frequency):
		updateStmt = """UPDATE agg_feeds
						   SET nextCheck = DATE_ADD(%s, INTERVAL %s MINUTE)
					     WHERE feed_id = %s"""
		self.cursor.execute (updateStmt, (mx.DateTime.gmt(), frequency, str (self.id)))
		
	def getUpdates (self):
		self.log.info ("Processing feed %s (%s)" % (str (self.id), self.url))
		lastUpdate = self.getLastUpdate ()
		channelDetails = None
		try:
			request = urllib2.Request (self.url)
			if (lastUpdate is not None):
				lastModifiedDate = lastUpdate.strftime("%a, %d %b %Y %H:%M:%S GMT")
				self.log.debug ("Performing HTTP Get with last modified of: %s" % lastModifiedDate)
				request.add_header ('If-Modified-Since', lastModifiedDate)
			try:
				request.add_header("User-Agent", USER_AGENT)
				reply = urllib2.urlopen (request)
				self.log.debug ("Retrieved RSS file, sending to RSSParser")
				try:
					parser = RSSParser.Parser (reply.read())
				except Exception, e:
					self.log.error ("Error during parsing: " + str (e))
					return
				if (parser.getStatus()):
					try:
						channelDetails = self.handleRSS (parser)
					except Exception, e:
						self.log.error ("Error handling RSS: " + str (e))
						return
					try:
						self.updateFeedDetails (channelDetails)
					except Exception, e:
						self.log.error ("Error updating feed details: " + str (e))
						return
				else:
					self.log.error ("Unable to parse retrieved RSS feed.")
			except urllib2.URLError, e:
				if (e.code == 304):
					self.log.info ("304 Returned - Feed not modified since last get")
				else:
					self.log.error ("HTTP Response Exception " + str (e))
		except Exception, e:
			self.log.error ("Error during feed retrieval: " + str (e))
					
	def getLastUpdate (self):
		stmt = "SELECT lastUpdate FROM agg_feed_details WHERE feed_id = %s" 
		self.cursor.execute (stmt, (str (self.id),))
		result = self.cursor.fetchone()
		if (result is None):
			self.log.debug ("There are currently no feed details for this feed, creating some.")
			crstmt = "INSERT INTO agg_feed_details (feed_id) values (%s)"
			self.cursor.execute (crstmt, (str (self.id),))
			return None
		if (result[0] is None):
			self.log.debug ("Last update is NULL!")
			return None
		return result[0]
		
	def updateFeedDetails (self, newDetails):
		stmt = """UPDATE agg_feed_details
					 SET title = %s
						,lastUpdate = %s
						,imageURL = %s
						,imageTitle = %s
						,imageLink = %s
						,channelLink = %s
						,channelDesc = %s
						,htmlChannelDesc = %s
						,language = %s
						,goodXML = %s
				   WHERE feed_id = %s"""
		if (newDetails ['title'] is not None):
			newTitle = htmlStrip.stripHTML (newDetails ['title'])
		else:
			newTitle = None
		if (newDetails ['channelDesc'] is not None):
			newDesc = htmlStrip.stripHTML (newDetails['channelDesc'])
		else:
			newDesc = None
													
		self.cursor.execute (stmt, (unicodeToDb (newTitle)
									 ,mx.DateTime.gmt()
									 ,newDetails['imageURL']
									 ,unicodeToDb (newDetails['imageTitle'])
									 ,newDetails['imageLink']
									 ,newDetails['channelLink']
									 ,unicodeToDb (newDesc)
									 ,unicodeToDb (newDetails['channelDesc'])
									 ,newDetails['language']
									 ,newDetails['goodXML']
									 ,str (self.id)))
	def handleRSS (self, parser):
		# Need to:
		# 1 - Pull out the channel details
		# 2 - Compare every article in the feed with ones we already have
		self.log.debug ("Looking at parsed data.")
		channelDetails = {}
		chanDets = parser.getChannelDetails()
		channelDetails ['goodXML'] = parser.getXMLStatus()
		channelDetails ['title'] = chanDets.get ('title', None)
		channelDetails ['channelLink'] = chanDets.get ('link', None)
		channelDetails ['language'] = chanDets.get ('language', None)
		channelDetails ['channelDesc'] = chanDets.get ('description', None)
		imageDets = chanDets.get ('image', {})
		channelDetails ['imageURL'] = imageDets.get ('url', None)
		channelDetails ['imageTitle'] = imageDets.get ('title', None)
		channelDetails ['imageLink'] = imageDets.get ('link', None)
		
		#self.log.debug ("Channel details are: " + str (channelDetails))
		# Get the build date in GMT
		lastBuildDate = parser.getBuildDate()
		# Do this outside the loop so that all articles found in this feed have the
		# same discovery date.
		discoveryDate = mx.DateTime.gmt()
		
		# Now pull out all the articles.
		checkForExisting = """SELECT article_id, title, link, htmlDescription FROM agg_articles WHERE feed_id = %s AND fingerPrint = %s"""
		totalExistingArticles = """SELECT COUNT(*) FROM agg_articles WHERE feed_id = %s"""
		updateExistingArticle = """UPDATE agg_articles 
									  SET title=%s
										 ,link=%s
										 ,description=%s
										 ,htmlDescription=%s
										 ,articleDate=%s
								   WHERE article_id=%s"""
		insertNewArticle = """INSERT INTO agg_articles
													(feed_id
													,articleDate
													,discoveryDate
													,title
													,description
													,htmlDescription
													,link
													,guid
													,fingerPrint
													,firstFeedFlag
													)
												VALUES
													(%s
													,%s
													,%s
													,%s
													,%s
													,%s
													,%s
													,%s
													,%s
													,%s
													)"""
		self.cursor.execute (totalExistingArticles, (self.id,))
		numExistingArticles = int (self.cursor.fetchone()[0])
		if (numExistingArticles == 0):
			self.log.info ("First ever read of this feed - marking articles as such after the first insert.")
			firstReadOfFeed = 1
		else:
			self.log.debug ("Old feed with a total of %s articles already read." % str (numExistingArticles))
			firstReadOfFeed = 0
		insertCount = 0
		recordCur = self.db.cursor()
		for article in parser.getItems():
			self.log.debug ("Checking for an existing article...")
			# If guid is present then use this as the fingerPrint, else use 
			# a checksum of the title and link
			self.log.debug ('Calculating finger print...')
			if (article.has_key ('guid')):
				self.log.debug ('Finger print is GUID')
				fingerPrint = article['guid']
				guidIsFingerPrint = 1
			else:
				self.log.debug ('Finger print is title/link')
				checkSum = md5.new()
				checkSum.update (ASCII_ENCODER (article.get ('title', ''), 'ignore')[0])
				checkSum.update (ASCII_ENCODER (article.get ('link', ''), 'ignore')[0])
				fingerPrint = checkSum.hexdigest()
				guidIsFingerPrint = 0
			
			self.log.debug ('Looking for aticle with fingerPrint of %s' % fingerPrint)
			self.cursor.execute (checkForExisting, (self.id, fingerPrint))
			self.log.debug ("Cursor execute completed.")
			alreadyPresent = 0
			if (article.has_key ('title')):
				self.log.debug ("Stripping new title of any HTML that might be present")
				newTitle = htmlStrip.stripHTML (article ['title'])
			else:
				newTitle = None
			self.log.debug ("Looking at each possible match for this article.")
			for possibleMatch in self.cursor.fetchall():
				oldID = possibleMatch[0]
				if (possibleMatch[1] is None):
					oldTitle = None
				else:
					oldTitle = dbToUnicode (possibleMatch[1])
				oldLink = possibleMatch[2]
				if (possibleMatch[3] is None):
					oldHTMLDescription = None
				else:
					oldHTMLDescription = dbToUnicode (possibleMatch[3])
				self.log.debug ("Finger-print match on article ID %s" % str (oldID))
				# GUID should always be unique, so only check for the checksum generating matches
				# on different titles/links
				if (not guidIsFingerPrint):
					# Convert new data to DB charset and back so that any replaced characters are matched
					if (dbToUnicode (unicodeToDb (newTitle)) != oldTitle) or (article.get ('link', None) != oldLink):
						self.log.warn ("FingerPrint found match, but title or link is different!")
					else:
						self.log.debug ("Article confirmed as already present.")
						alreadyPresent = 1
				else:
					self.log.debug ("GUID Match found - aritlce already present")
					alreadyPresent = 1
				
				# Check to see whether any found article should be updated.
				if (alreadyPresent):
					self.log.debug ("Checking for a possible update to the article.")
					# Convert new data to DB charset and back so that any replaced characters are matched
					if (dbToUnicode (unicodeToDb (newTitle)) != oldTitle) or (dbToUnicode (unicodeToDb (article.get ('description', None))) != oldHTMLDescription):
						self.log.info ('Article has been updated since last read, updating database with new content.')
						if (article.has_key ('description')):
							newDesc = htmlStrip.stripHTML (article['description'])
						else:
							newDesc = None
						recordCur.execute (updateExistingArticle, (unicodeToDb (newTitle)
																	, article.get ('link', None)
																	, unicodeToDb (newDesc)
																	, unicodeToDb (article.get ('description', None))
																	,article.get ('pubDate', lastBuildDate)
																	, possibleMatch[0]))
					self.log.debug ("Checking complete.")
									
			if (not alreadyPresent):
				self.log.info ("Feed contained a new article - adding to database")
				if (article.has_key ('description')):
					newDesc = htmlStrip.stripHTML (article['description'])
				else:
					newDesc = None
				
				if (firstReadOfFeed and insertCount > 0):
					# Mark these with the first read flag
					firstFeedFlag = 'Y'
				else:
					firstFeedFlag = 'N'

				recordCur.execute (insertNewArticle, (str (self.id)
													 ,article.get ('pubDate', lastBuildDate)
													 ,discoveryDate
													 ,unicodeToDb(newTitle)
													 ,unicodeToDb(newDesc)
													 ,unicodeToDb(article.get ('description', None))
													 ,article.get ('link', None)
													 ,article.get ('guid', None)
													 ,fingerPrint
													 ,firstFeedFlag))
				self.log.debug ("Inserted.")
				insertCount += 1
			
		recordCur.close()
		return channelDetails

aggregator = Aggregator()
aggregator.runAggregator()
